﻿/*
 ExtractWav:
    Extracts the WAV portion of an SCI0 sound resource input file, either a patch resource or a packed one (RESOURCE.XXX).
    Uncompressed or LZW compressed files are supported.  Huffman decompression not implemented.
 
 Do whatever you want with this file, except remove my name.
 Jeremiah Nellis, nellisjp@hotmail.com

 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;


/*
 * 1) Check input sound file, ensure it contains 84 00 02 (02 means it's got a wav file in it
 * 
 * 2) WAV header format:
 *    WAV resources start with: 1A 00 01 (3 bytes)
 *    Then: 00 00 00 00 00 00 00 00 00 00 00 (11 bytes) - SQ3, all sound.xxx
 *          00 00 F0 BF F0 A3 F0 97 F0 BF F0 (11 bytes) - KQ1, sound.015
 *    The sample size: XX XX (2 bytes)
 *    Then: FF 00 FF FF 2C (5 bytes)
 *    Then: 00 00 00 00 00 00 00 00 00 00 00 (11 bytes)
 *    The resource size: XX XX (2 bytes)
 *    Then: 00 00 00 00 00 00 00 00 00 00 (10 bytes)
 *    --------
 *    44 byte header
 */

namespace ConsoleApplication1
{
    class ExtractWav
    {
        public static string pathSource;
        public static string outputDir;
        public static int extractCount = 0;


        
        /* 9-12 bit LZW encoding */
        public static byte[] decrypt1(byte[] src, int length, int complength)
             /* Doesn't do length checking yet */
        {
	        /* Theory: Considering the input as a bit stream, we get a series of
	        ** 9 bit elements in the beginning. Every one of them is a 'token'
	        ** and either represents a literal (if < 0x100), or a link to a previous
	        ** token (tokens start at 0x102, because 0x101 is the end-of-stream
	        ** indicator and 0x100 is used to reset the bit stream decoder).
	        ** If it's a link, the indicated token and the character following it are
	        ** placed into the output stream. Note that the 'indicated token' may
	        ** very well consist of a link-token-plus-literal construct again, so
	        ** it's possible to represent strings longer than 2 recursively.
	        ** If the maximum number of tokens has been reached, the bit length is
	        ** increased by one, up to a maximum of 12 bits.
	        ** This implementation remembers the position each token was print to in
	        ** the output array, and the length of this token. This method should
	        ** be faster than the recursive approach.
	        */

            byte[] dest = new byte[length];
            //Console.WriteLine(src.Length);

	        ushort bitlen = 9; /* no. of bits to read (max. 12) */
	        ushort bitmask = 0x01ff;
	        ushort bitctr = 0; /* current bit position */
	        ushort bytectr = 0; /* current byte position */
	        ushort token; /* The last received value */
	        ushort maxtoken = 0x200; /* The biggest token */

            ushort[] tokenlist = new ushort[4096]; /* pointers to dest[] */
	        ushort[] tokenlengthlist = new ushort[4096]; /* char length of each token */
	        ushort tokenctr = 0x102; /* no. of registered tokens (starts here)*/

	        ushort tokenlastlength = 0;

	        ushort destctr = 0;

	        while (bytectr < complength) {

		        ushort tokenmaker = (ushort)(src[bytectr++] >> bitctr);
		        if (bytectr < complength)
			        tokenmaker |= (ushort)(src[bytectr] << (8-bitctr));
		        if (bytectr+1 < complength)
			        tokenmaker |= (ushort)(src[bytectr+1] << (16-bitctr));

		        token = (ushort)(tokenmaker & bitmask);

		        bitctr += (ushort)(bitlen - 8);

		        while (bitctr >= 8) {
			        bitctr -= 8;
			        bytectr++;
		        }

		        if (token == 0x101) return dest; /* terminator */
		        if (token == 0x100) { /* reset command */
			        maxtoken = 0x200;
			        bitlen = 9;
			        bitmask = 0x01ff;
			        tokenctr = 0x0102;
		        } else {

			        {
				        int i;

				        if (token > 0xff) {
				          if (token >= tokenctr)
				            {

				              //fprintf(stderr, "decrypt1: Bad token %x!\n", token);

				              /* Well this is really bad  */
				              /* May be it should throw something like SCI_ERROR_DECOMPRESSION_INSANE */
				            } else
				              {
					        tokenlastlength = (ushort)(tokenlengthlist[token]+1);
					        if (destctr+tokenlastlength>length)
					          {


					            /* For me this seems a normal situation, It's necessary to handle it*/
					            //printf ("decrypt1: Trying to write beyond the end of array(len=%d, destctr=%d, tok_len=%d)!\n",
						        //    length, destctr, tokenlastlength);


					            i = 0;
					            for (; destctr<length; destctr++) {
					              dest[destctr++] = dest [tokenlist[token]+i];
					              i++;
					            }
					          } else
					        for (i=0; i< tokenlastlength; i++) {
						        dest[destctr++] = dest[tokenlist[token]+i];
					        }
				              }
				        } else {
					        tokenlastlength = 1;
				          if (destctr >= length)
				            {

				              Console.WriteLine("decrypt1: Try to write single byte beyond end of array!\n");

				            } else
					        dest[destctr++] = (byte)token;
				        }

			        }

			        if (tokenctr == maxtoken) {
				        if (bitlen < 12) {
					        bitlen++;
					        bitmask <<= 1;
					        bitmask |= 1;
					        maxtoken <<= 1;
				        } else continue; /* no further tokens allowed */
			        }

			        tokenlist[tokenctr] = (ushort)(destctr-tokenlastlength);
			        tokenlengthlist[tokenctr++] = tokenlastlength;

		        }

	        }

	        return dest;

        }

        private static string hex2binary(string hexvalue)
        {
            string binaryval = "";
            binaryval = Convert.ToString(Convert.ToInt32(hexvalue, 16), 2);
            return binaryval;
        }

        public static void processResource(String hexString)
        {
            String hexResNum = "";
            int resLengthComp = 0;
            int resLengthDecomp = 0;
            int resType = 0;
            int resNum = -1;  // init for patch resource
            int compType = 0;
            String currentRes = "";

            // This is a SOUND.XXX (patch) file
            if (hexString.Substring(0, 4) == "8400")
            {
                extractWav(hexString, resNum);
                return;
            }

            // This is a RESOURCE.XXX file
            while (hexString.Length > 0)
            {

                // Get the first 16 characters (8 bytes) of the header
                hexResNum = hexString.Substring(0, 4);
                resLengthComp = Convert.ToInt32(hexString.Substring(4, 2), 16) + Convert.ToInt32(hexString.Substring(6, 2), 16) * 256 + 4;
                resLengthDecomp = Convert.ToInt32(hexString.Substring(8, 2), 16) + Convert.ToInt32(hexString.Substring(10, 2), 16) * 256 + 4;
                resType = Convert.ToInt32(hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(0, 5), 2);
                resNum = Convert.ToInt32(hexString.Substring(0, 2), 16)
                            + Convert.ToInt32(hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(5, 3), 2) * 256;

                compType = Convert.ToInt32(hexString.Substring(12, 2), 16);

                // Split into two pieces, this resource and the rest of the file
                try
                {
                    currentRes = hexString.Substring(16, resLengthComp * 2 - 16);
                    hexString = hexString.Substring(resLengthComp * 2);
                }
                catch
                {
                    Console.WriteLine("Non SCI0 resource or resource patch file.  Exiting.");
                    Environment.Exit(2); // Exit with error
                }

                // if this is a sound resource
                if (resType == 4)
                {
                    // no compression
                    if (compType == 0)
                    {
                        extractWav(currentRes, resNum);
                    }

                    // LZW compression
                    if (compType == 1)
                    {
                        byte[] destination = decrypt1(HexStringToBytes(currentRes), resLengthDecomp - 4, resLengthComp - 8);
                        String decompressedRes = BytesToHexString(destination);
                        extractWav(decompressedRes, resNum);
                    }

                    // Huffman
                    if (compType == 2)
                    {
                        Console.WriteLine("Skipping Huffman compressed resource (not currently supported).");
                    }
                }

            }


        }

        private const string hexDigits = "0123456789ABCDEF";

        public static string BytesToHexString(byte[] bytes)
        {
            StringBuilder sb = new StringBuilder(bytes.Length * 2);
            foreach (byte b in bytes)
            {
                sb.AppendFormat("{0:X2}", b);
            }
            return sb.ToString();
        }

        public static byte[] HexStringToBytes(string str)
        {
            // Determine the number of bytes
            byte[] bytes = new byte[str.Length >> 1];
            for (int i = 0; i < str.Length; i += 2)
            {
                int highDigit = hexDigits.IndexOf(Char.ToUpperInvariant(str[i]));
                int lowDigit = hexDigits.IndexOf(Char.ToUpperInvariant(str[i + 1]));
                if (highDigit == -1 || lowDigit == -1)
                {
                    throw new ArgumentException("The string contains an invalid digit.", "s");
                }
                bytes[i >> 1] = (byte)((highDigit << 4) | lowDigit);
            }
            return bytes;
        }


        public static void extractWav(String data, int resNum) {

            string fileSansExt = Path.GetFileNameWithoutExtension(pathSource);
            string fileExt = Path.GetExtension(pathSource);

            Regex rx = new Regex(@"(1A0001    
                                    [A-F0-9]{22}) # ???
                                    ([A-F0-9]{4}) # sample size
                                    (FF00FFFF2C
                                    [0]{22})
                                    ([A-F0-9]{4}) # resource size
                                    ([0]{20})
                                   ", RegexOptions.IgnorePatternWhitespace | RegexOptions.Singleline);

            MatchCollection MatchList = rx.Matches(data);

            // There should be only one match, one WAV possible per sound resource.
            foreach (Match match in MatchList)
            {
                int sampleRate = 0;
                int resourceSize = 0;
                string pathNew = "";
                string pathStripped = "";

  
                if (resNum == -1)
                {
                    pathNew = outputDir + '\\' + fileSansExt + '_' + fileExt.Remove(0, 1)  + ".wav";
                    pathStripped = outputDir + "\\stripped\\" + fileSansExt + fileExt;
                }
                else
                {
                    pathNew = outputDir + '\\' + fileSansExt + '_' + fileExt.Remove(0, 1) + '_'
                            + Convert.ToString(resNum).PadLeft(3,'0') + ".wav";

                    pathStripped = outputDir + "\\stripped\\" + fileSansExt + '_' + fileExt.Remove(0, 1) + "." 
                                 + Convert.ToString(resNum).PadLeft(3, '0');
                }


                // Get the beginning of the resource (the non-digital part) & write it out
                String soundRes = data.Substring(0, match.Index);
                
                // Make sure the data string is prepended with an 8400 (sound patch resource designator)
                // and turn off the digital resource indicator (replace the 2 with
                if(soundRes.Substring(0,4) != "8400") {
                    soundRes = "00" + soundRes.Substring(2, soundRes.Length - 2);
                    soundRes = "8400" + soundRes;
                }
                byte[] soundResbytes = HexStringToBytes(soundRes);

                // Write out the stripped file

                try
                {
                    System.IO.FileInfo fileStripped = new System.IO.FileInfo(pathStripped);
                    fileStripped.Directory.Create(); // If the directory already exists, this method does nothing.
                    FileStream fsNewStripped = new FileStream(pathStripped, FileMode.Create, FileAccess.Write);
                    fsNewStripped.Write(soundResbytes, 0, soundResbytes.Length);
                    fsNewStripped.Close();
                    extractCount++;
                }
                catch
                {
                    Console.WriteLine("Could not create output file!");
                    Environment.Exit(2);
                }

                foreach (Capture capture in match.Captures)
                {
                    Group GroupCurrent;
                    for (int i = 1; i < match.Groups.Count; i++)
                    {
                        GroupCurrent = match.Groups[i];

                        if (i == 2)
                        {
                           sampleRate = System.Int32.Parse(GroupCurrent.Value.Substring(2, 2) + GroupCurrent.Value.Substring(0, 2), 
                                                           System.Globalization.NumberStyles.AllowHexSpecifier);
                        }

                        if (i == 4)
                        {
                           resourceSize = System.Int32.Parse(GroupCurrent.Value.Substring(2, 2) + GroupCurrent.Value.Substring(0, 2),
                                                             System.Globalization.NumberStyles.AllowHexSpecifier);
                        }

                    }

                    // Load up a string with our header
                    String headerstr1 = "52494646"; // RIFF

                    String headerstr2 = "57415645666D7420"; // WAVEfmt
                    headerstr2 += "10000000"; // Subchunk1Size (16 for PCM)
                    headerstr2 += "0100"; // Audio Format (PCM)
                    headerstr2 += "0100"; // Number of channels (1 for mono)

                    String headerstr3 = "0100"; // Block size
                    headerstr3 += "0800";   // Bits per sample (8)
                    headerstr3 += "64617461"; // data subchunk
                    
                    byte[] temp = BitConverter.GetBytes(resourceSize + 36); // Resource size + 36 (total size of file following this number)
                    Array.Reverse(temp);
                    int test = BitConverter.ToInt32(temp, 0);
                    String resourceSizePlus36Hex = test.ToString("X").PadLeft(8, '0');

                    temp = BitConverter.GetBytes(resourceSize); // Resource size
                    Array.Reverse(temp);
                    test = BitConverter.ToInt32(temp, 0);
                    String resourceSizeHex = test.ToString("X").PadLeft(8, '0');

                    temp = BitConverter.GetBytes(sampleRate); // sample rate
                    Array.Reverse(temp);
                    test = BitConverter.ToInt32(temp, 0);
                    String sampleRateHex = test.ToString("X").PadLeft(8, '0');
                    
                    byte[] resourceSize1Bytes = HexStringToBytes(resourceSizePlus36Hex);
                    byte[] resourceSize2Bytes = HexStringToBytes(resourceSizeHex);
                    byte[] sampleRateBytes = HexStringToBytes(sampleRateHex);
                    byte[] header1bytes = HexStringToBytes(headerstr1);
                    byte[] header2bytes = HexStringToBytes(headerstr2);
                    byte[] header3bytes = HexStringToBytes(headerstr3);
                    
                    // Write out file
                    Console.WriteLine("Extracting: " + pathNew);
                    System.IO.FileInfo file = new System.IO.FileInfo(pathNew);
                    file.Directory.Create(); // If the directory already exists, this method does nothing.
                    FileStream fsNew = new FileStream(pathNew, FileMode.Create, FileAccess.Write);

                    fsNew.Write(header1bytes, 0, header1bytes.Length);
                    fsNew.Write(resourceSize1Bytes, 0, resourceSize1Bytes.Length);
                    fsNew.Write(header2bytes, 0, header2bytes.Length);
                    fsNew.Write(sampleRateBytes, 0, sampleRateBytes.Length);
                    fsNew.Write(sampleRateBytes, 0, sampleRateBytes.Length); // Byte rate, which in our case matches the sample rate
                    fsNew.Write(header3bytes, 0, header3bytes.Length);
                    fsNew.Write(resourceSize2Bytes, 0, resourceSize2Bytes.Length);

                    byte[] wavdata = HexStringToBytes(data.Substring(capture.Index + 88, resourceSize * 2));
                    fsNew.Write(wavdata, 0, wavdata.Length);

                    fsNew.Close();

                }

            }

         }

        static int Main(string[] args)
        {

            /*
            pathSource = @"c:\temp\resources\kq1sci\resource.001";
            //pathSource = @"c:\sci\sounds\trekdoor.000";
            outputDir = @"c:\temp\extract";
            outputDir = Path.GetFullPath(outputDir);
            */

            
            // Test if input arguments were supplied:
            if (args.Length != 2)
            {
                Console.WriteLine("ExtractWav");
                Console.WriteLine("Version 1.0");
                Console.WriteLine("By Jeremiah Nellis");
                Console.WriteLine();
                Console.WriteLine("Extracts WAV files from patch or packed SCI0 files");
                Console.WriteLine("Usage: ExtractWav.exe <inputResource> <outputDirectory>");
                Console.WriteLine("Examples:");
                Console.WriteLine(@"  ExtractWav.exe resource.001 c:\fooDir");
                Console.WriteLine(@"  ExtractWav.exe c:\randomdir\sound.001 .\barDir");
                return 1; 
            }

         
            pathSource = args[0];
            outputDir = Path.GetFullPath(args[1]);
            byte[] _bytes = null;

            try
            {
                 _bytes = File.ReadAllBytes(pathSource);
            }
            catch
            {
                Console.WriteLine("Could not open input file!");
                return 2;
            }
            String data = BytesToHexString(_bytes);
            processResource(data);

            if (extractCount > 0)
            {
                Console.WriteLine("WAV data succesfully extracted.");
            }
            else
            {
                Console.WriteLine("No suitable WAV data found to extract.");
            }

            return 0; 
        }

    }

}
