﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;

namespace SndPrf
{
    class SnfPrf
    {

        /*
         *  SCI Sound Resource Profiler
         *  By Ravi I.
         *
         *  This program takes a sound file from the command line and lists all of
         *  the events. The header is currently ignored because I have no idea what
         *  to do with it yet.
         * 
         *  Ported to C# by Jeremiah Nellis - June 17th, 2012
        */


        static int Main(string[] args)
        {
           BinaryReader infile = null;

           // if the user did not pass arguments, display help
           if ( args.Length != 1 )
           {
              DisplayHelp();
              return 1;
           }

           // open the file and quit on error

           try
           {
               infile = new BinaryReader(File.Open(args[0], FileMode.Open));
           }
           catch
           {
               ErrorExit("Error: Unable to open sound file for input");
           }


           // check for a valid sound resource
           if ( GetIdentifier(ref infile) != 0x0084 )
              ErrorExit ("Error: Unrecognized file format");

           // display the sound events
           DumpEvents (ref infile);

           // close the file and exit
           infile.Close();
           return 0;
        }


        //-------------------------------------------------------------------------
        //   DisplayHelp
        //   Called when the program is run without command-line arguments
        //   Displays a help screen with program usage

        static void DisplayHelp()
        {
           Console.WriteLine("SCI Sound Resource Profiler");
           Console.WriteLine("version 0.02");
           Console.WriteLine("By Ravi Iyengar");
           Console.WriteLine("Ported to C# by Jeremiah Nellis - June 17th, 2012");
           Console.WriteLine("Usage:");
           Console.WriteLine("   sndprf <infile>");
        }


        //-------------------------------------------------------------------------
        //   ErrorExit
        //   Called when the program encounters an error and must quit
        //   Displays an error message then exits

        static void ErrorExit(string message)
        {
            Console.WriteLine(message);
            Environment.Exit(1);
        }


        //-------------------------------------------------------------------------
        //   GetIdentifier
        //   Called with the input file handle to check for a sound resource
        //   Returns the first two bytes of the file as a short int

        static short GetIdentifier(ref BinaryReader fileh)
        {
            fileh.BaseStream.Position = 0;
            byte[] id = fileh.ReadBytes(2);
            return (short)((int)id[0] + (int)id[1]);
        }


        //-------------------------------------------------------------------------
        //   DumpEvents
        //   Called with the input file handle to list sound events
        //   Displays one event per line

        static void DumpEvents (ref BinaryReader fileh)
        {
           byte foo, statusbyte = 0, p1, p2, channel;
           short bar, eventtime;
           string command;


           // skip the file header
           fileh.BaseStream.Position = 35;

           // loop until the end of the file
           while ( (eventtime = ReadChar(ref fileh)) != -1 )  // get the delta time
           {

              // get the next byte
              if ( (bar = ReadChar(ref fileh)) == -1 )
                 ErrorExit ("Error: Unexpected end of file");
              foo = (byte)bar;

              // check for "running status byte" by checking bit 7
              if ( Convert.ToBoolean(foo & 0x80) )
                 statusbyte = foo;              // this is a command
              else
                  fileh.BaseStream.Position--;   // rewind to read it as a paramater
                 

              // interpret the command
              switch ( statusbyte & 0xF0 )
              {
                 case 0x80:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p2 = (byte)bar;
                    command = "8x: Note Off    Stop note " + p1 + " with " + p2 + " velocity on channel " + channel;
                    break;

                 case 0x90:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p2 = (byte)bar;
                    if ( p2 > 0 )
                       command = "9x: Note On     Play note " + p1 + " with velocity " + p2 + " on channel " + channel;
                    else
                       command = "9x: Note On     Stop note (0 velocity) " + p1 + " on channel " + channel;
                    break;

                 case 0xA0:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p2 = (byte)bar;
                    command = "Ax: Pressure    After-touch key pressure for pitch " + p1 + " to " + p2 + " on channel " + channel;
                    break;

                 case 0xB0:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p2 = (byte)bar;
                    command = "Bx: Control     Set control " + p1 + " to " + p2 + " on channel " + channel;
                    break;

                 case 0xC0:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    command = "Cx: Program     Set program to " + p1 + " on channel " + channel;
                    break;

                 case 0xD0:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    command = "Dx: Pressure    After-touch channel pressure to " + p1 + " on channel " + channel;
                    break;

                 case 0xE0:
                    channel = (byte)(statusbyte & 0x0F);
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p1 = (byte)bar;
                    if ( (bar = ReadChar (ref fileh)) == -1)
                       ErrorExit ("Error: Unexpected end of file");
                    p2 = (byte)bar;
                    command = "Ex: Pitch       Set pitch wheel to " + (p1 | (p2 << 7)) + " on channel " + channel;
                    break;

                 case 0xF0:
                    if ( statusbyte == 0xFC )
                       command = "FC: End Sequence";
                    else
                    {
                        command = "Error: Unrecognized command " + statusbyte + "";
                       ErrorExit (command);
                    }
                    break;

                 default:
                    command = "Error: Invalid command " + statusbyte + "";
                    ErrorExit (command);
                    break;
              }

              Console.WriteLine("Delta time: {0}    {1}", eventtime.ToString().PadLeft(3, ' '), command);
           }

           if ( Convert.ToBoolean(statusbyte != 0xFC) )
              ErrorExit ("Error: Unexpected end of file");
        }


        //-------------------------------------------------------------------------
        //   ReadChar
        //   Called with the input file handle to read a single byte
        //   Returns the value or -1 on error

        static short ReadChar(ref BinaryReader fileh)
        {
            byte? charbuf;

            try
            {
                charbuf = fileh.ReadBytes(1)[0];
            } catch{
                return -1;
            }

            return (short)charbuf;
        }
    }
}
