﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;
using System.Xml.Linq;

namespace SoundToMidi
{
    class SoundToMidi
    {

        /*****************************************************
        * SND2MIDI version 0.04.3 (current)               
        * 
        * Change history:
        * ---------------
        *   v 0.04.0 - ??/??/???? - Original application programmed by Ravi I.
        *   v 0.04.1 - Updated to be Win32 compatible by Brandon Blume on April 3, 2012    
        *   v 0.04.2 - Ported to C# by Jeremiah Nellis on June 17, 2012 
        *   v 0.04.3 - Added support for dumping voice/device sound data into output XML file
        ******************************************************/

        static int Main(string[] args)
        {
           ulong tracklen;
           BinaryReader infile = null;
           BinaryWriter outfile = null;
           int instrument = -1;
           byte foo;
           string xmlHeaderFile = null;
           string inputFileName = null;

        
           // check for command-line arguments
           if ( args.Length < 2 )
           {
              DisplayHelp();
              return 0;
           }

           // open the input and output files
           try
           {
               infile = new BinaryReader(File.Open(args[0], FileMode.Open));
               inputFileName = args[0];
           }
           catch
           {
               ErrorExit("Error: Unable to open sound file for input\n");
           }

           try
           {
               outfile = new BinaryWriter(File.Open(args[1], FileMode.Create));
           }
           catch
           {
               ErrorExit("Error: Unable to open midi file for output\n");
           }


           // interpret flags
           for (int lcv = 2; lcv < args.Length; lcv++)
           {
               if (args[lcv][0] == '-')
                   switch (args[lcv][1])
                   {
                       case 'I':
                           instrument = Convert.ToInt16(args[lcv + 1]);
                           break;

                       case 'X':
                           xmlHeaderFile = args[lcv + 1];
                           break;

                       default:
                           Console.Write("Unknown flag {0}\n", args[lcv]);
                           break;
                   }
           }

           // check for a valid sound resource
           if ( GetIdentifier(ref infile) != 0x0084 )
              ErrorExit ("Error: Unrecognized file format\n");

           // read the SOUND header & write out to XML information file
           if (xmlHeaderFile != null)
           {
               WriteXMLHeaderFile(ref infile, ref inputFileName, ref xmlHeaderFile);
           }

           // write the MIDI header and track header
           WriteMThd(ref outfile);
           WriteMTrk(ref outfile);

           // copy events
           tracklen = CopyEvents(ref infile, ref outfile, instrument);

           // set the track length (written MSB first)           
           outfile.BaseStream.Position = 18;
           foo = (byte)(tracklen >> 24);
           outfile.Write(foo);
           foo = (byte)(tracklen >> 16);
           outfile.Write(foo);
           foo = (byte)(tracklen >> 8);
           outfile.Write(foo);
           foo = (byte)(tracklen);
           outfile.Write(foo);
       
           // close the open files
           infile.Close();
           outfile.Close();

           return 0;
        }


        //-------------------------------------------------------------------------
        // DisplayHelp
        // Called when the program is run without argumenmts
        // Displays help information

        static void DisplayHelp()
        {
           Console.Write("\nSCI Sound to MIDI Converter\n");
           Console.Write("version 0.04.3\n");
           Console.Write("By Ravi I.\n");
           Console.Write("Updated to Win32 by Brandon Blume April 3, 2012\n");
           Console.Write("Ported to C# by Jeremiah Nellis June 17, 2012\n");
           Console.Write("    Modified to output voice/device data, JN - June 23, 2012\n");
           Console.Write("Usage:\n");
           Console.Write("   snd2midi <soundfile> <midifile> [flags]\n\n");
           Console.Write("<soundfile> is the filename of an extracted sound resource\n");
           Console.Write("<midifile> is the output file\n\n");
           Console.Write("Flags must come after the filenames if included. A flag\n");
           Console.Write("must be separated from its paramater with a space.\n");
           Console.Write("   -I n     - set MIDI program to n\n");
           Console.Write("   -X <filename> - Dumps device sound header information to XML file.\n");
        }


        //-------------------------------------------------------------------------
        //   ErrorExit
        //   Called when the program encounters an error and must quit
        //   Displays an error message then exits

        static void ErrorExit (string message)
        {
           Console.Write(message);
           Environment.Exit(1);
        }

        //-------------------------------------------------------------------------
        //   GetIdentifier
        //   Called with the input file handle to check for a sound resource
        //   Returns the first two bytes of the file as a short int

        static short GetIdentifier (ref BinaryReader fileh)
        {
           fileh.BaseStream.Position = 0;
           byte[] id = fileh.ReadBytes(2);
           return (short)((int)id[0] + (int)id[1]);
        }


        //-------------------------------------------------------------------------
        //   ReadChar
        //   Called with the input file handle to read a single byte
        //   Returns the value or -1 on error

        static short ReadChar (ref BinaryReader fileh)
        {
           byte? charbuf;

           charbuf = fileh.ReadBytes(1)[0];
           if(charbuf == null) {
              return -1;
           }

           return (short)charbuf;
        }


        //-------------------------------------------------------------------------
        //   WriteVL
        //   Called with the output file handle and a data byte
        //   Writes the variable-legth value and returns no. of bytes written

        static char WriteVL (ref BinaryWriter fileh, char data)
        {
           char num = (char)0, topbyte = (char)0x81;
           if (Convert.ToBoolean(data & 0x80))
           {
              data &= (char)0x7F;

              fileh.Write(topbyte);
              num = (char)1;
           }
           fileh.Write(data);
           return ++num;
        }


        //-------------------------------------------------------------------------
        //   GetParam
        //   Called with the input handles
        //   Gets a command paramater

        static char GetParam (ref BinaryReader fileh)
        {
           short? foo;

           if ((foo = fileh.ReadBytes(1)[0]) == null)
              ErrorExit ("Error: Unexpected end of file\n");
           if ( Convert.ToBoolean(foo & 0x80) )
              Console.Write("Warning: Paramater with bit 7 set\n");
           return (char)foo;
        }



        //-------------------------------------------------------------------------
        //   CopyEvents
        //   Called with the input and output file handles plus special modifiers
        //   Copies the sound events returning the number of bytes written

        static ulong CopyEvents (ref BinaryReader inf, ref BinaryWriter outf, int MIDIProg)
        {
           ulong byteswritten = 0;
           byte statusbyte, laststatus = 0, p1, p2;
           short? bar, eventtime;

           // skip the file header
           inf.BaseStream.Position = 35;
            /*
           if ( lseek(inf, 0x23, SEEK_SET) == -1 )
              ErrorExit ("Error: Unable to seek to sound events\n");
           */

           // loop until the end of the file
           while ( (eventtime = inf.ReadBytes(1)[0]) != null )  // get the delta time
           {
              // get the next byte
              if ( (bar = inf.ReadBytes(1)[0]) == null )
                 ErrorExit ("Error: Unexpected end of file\n");
              statusbyte = (byte)bar;

              // check for running status byte by checking bit 7
              if ( !(Convert.ToBoolean(statusbyte & 0x80)) )
              {
                 statusbyte = laststatus;
                 inf.BaseStream.Position--;   // rewind to read it as a paramater
              }

              // copy the command
              switch ( statusbyte & 0xF0 )
              {
                 // these 2-paramater commands can be directly copied
                 case 0x80:    // note off
                 case 0x90:    // note on
                 case 0xA0:    // key pressure
                 case 0xB0:    // control change
                 case 0xE0:    // pitch wheel
                    p1 = (byte)GetParam(ref inf);
                    p2 = (byte)GetParam(ref inf);
                    byteswritten += WriteVL(ref outf, (char)eventtime);
                    if ( statusbyte != laststatus )
                    {
                       outf.Write(statusbyte);
                       byteswritten++;
                    }
                    outf.Write(p1);
                    outf.Write(p2);
                    byteswritten += 2;
                    break;

                 // these 1-paramater commands can be directly copied
                 case 0xD0:    // channel pressure
                    p1 = (byte)GetParam(ref inf);
                    byteswritten += WriteVL(ref outf, (char)eventtime);
                    if ( statusbyte != laststatus )
                    {
                       outf.Write(statusbyte);
                       byteswritten++;
                    }
                    outf.Write(p1);
                    byteswritten += 1;
                    break;

                 // the rest are special case commands

                 // check for set MIDI instrument
                 case 0xC0:
                    p1 = (byte)GetParam(ref inf);
                    if ( MIDIProg != -1 )
                       p1 = (byte)MIDIProg;
                    byteswritten += WriteVL(ref outf, (char)eventtime);
                    if ( statusbyte != laststatus )
                    {
                       outf.Write(statusbyte);
                       byteswritten++;
                    }
                    outf.Write(p1);
                    byteswritten += 1;
                    break;

                 // the end track command
                 case 0xF0:
                    if ( statusbyte != 0xFC )
                       ErrorExit ("Error: Unrecognized Fx command\n");
                    else
                       {
                          byteswritten += WriteVL(ref outf, (char)eventtime);
                          statusbyte = 0xFF;
                          p1 = 0x2F;
                          p2 = 0x00;
                          outf.Write(statusbyte);
                          outf.Write(p1);
                          outf.Write(p2);
                          byteswritten += 3;
                          return byteswritten;
                       }
                    break;

                 default:
                    ErrorExit ("Error: Invalid command\n");
                    break;
              }

              laststatus = statusbyte;
           }

           Console.Write("Warning: Unexpected end of file\n");

           return byteswritten;
        }



        //-------------------------------------------------------------------------
        //   WriteMThd
        //   Called with the handle a new MIDI file
        //   Writes the header

        static void WriteMThd (ref BinaryWriter outf)
        {
           //  MThd\x00\x00\x00\x06\x00\x00\x00\x01\x00\x20
           var header = new byte[] {0x4D, 0x54, 0x68, 0x64, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x20 };
           outf.Write(header);
        }



        //-------------------------------------------------------------------------
        //   WriteMTrk
        //   Called with the handle a new MIDI file
        //   Writes the track header

        static void WriteMTrk (ref BinaryWriter outf)
        {
           // MTrk\x00\x00\x00\x00
           var header = new byte[] { 0x4D, 0x54, 0x72, 0x6B, 0x00, 0x00, 0x00, 0x00 };
           outf.Write(header);
        }

        static void WriteXMLHeaderFile(ref BinaryReader fileh, ref string inputFileName, ref string xmlHeaderFile)
        {
          
            // set file to beginning of voice/channel data
            fileh.BaseStream.Position = 3;
            byte[] header = fileh.ReadBytes(32);

            var voices = header
                    .Where((elem, idx) => idx % 2 == 0);

            var devices = header
                    .Skip(1)
                    .Where((elem, idx) => idx % 2 == 0);

            XDocument doc =
              new XDocument(
                new XElement("file",
                  new XElement("name", new XAttribute("filename", inputFileName)),
                  new XElement("channels", from i in Enumerable.Range(0, 16)
                                            let num = new XAttribute("number", String.Format("{0}", i))
                                            let voice = new XAttribute("voices", voices.ElementAt(i))
                                            let device = new XAttribute("devices", devices.ElementAt(i))
                                            select new XElement("channel", num, voice, device)
                              )
                  )
                  
             
              );

            doc.Save(xmlHeaderFile);
        }

    }

}

